#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <mosquitto.h>
#include "isec.h"

#define demo_log_error(...)    do { printf("!!! Error on line %d: ", __LINE__); printf(__VA_ARGS__); printf("\n"); } while (0)

char *output_topic = NULL;

static char *demo_scanf(const char *fmt, ...);
static int demo_main(char *base64_dataT, char *base64_plain, char * plainP);

void on_message(struct mosquitto *mosq, void *userdata, const struct mosquitto_message *message) {
    if(message->payloadlen){
        char *base64_plainS  = (char *) calloc(strlen((char *)message->payload) + 1,sizeof(char));
        memcpy(base64_plainS,(char *)message->payload,strlen((char *)message->payload));

        char *base64_plainP  = (char *) calloc(4096,sizeof(char));
        char *plainP  = (char *) calloc(4096,sizeof(char));
        if (base64_plainP != NULL && plainP != NULL ) {
            // base64_plainP = malloc(4096);
            int rv = isec_err_fail;
            rv = demo_main(base64_plainS, base64_plainP, plainP);
            // isec_global_cleanup();
            if (isec_err_none == rv) {
                int msg_id = 0;
                int last_slash_index = -1;
                int len = strlen(message->topic);

                int i;
                for (i = len - 1; i >= 0; i--) {
                    if (message->topic[i] == '/') {
                        last_slash_index = i;
                        break;
                    }
                }

                if (last_slash_index != -1) {
                    // 输出斜杠后面的子字符串
                    // Publish a message
                    char *sta_start = &message->topic[last_slash_index + 1]; // 定位到 HB01 的开始位置
                    // char *sta_start = strrchr(message->topic, "/") + 0; // 定位到 HB01 的开始位置

                    if (sta_start != NULL) {
                        // 复制 output_topic 的内容到 output_topic_tem
                        char output_topic_tem[strlen(output_topic) + 1]; // +1 用于存放字符串结尾的 null 字符
                        strcpy(output_topic_tem, output_topic);
                        // 复制 HB01 到 topic2 中 {sta} 的位置
                        char *sta_placeholder = strstr(output_topic_tem, "{sta}");
                        if (sta_placeholder != NULL) {
                            strcpy(sta_placeholder, sta_start);
                        }
                        mosquitto_publish(mosq, &msg_id, output_topic_tem, strlen(plainP), plainP, 0, false);
                    }
                }
            }
        } else {
            // 处理内存分配失败的情况
            fprintf(stderr, "Memory allocation failed.\n");
        }
        printf("input_topic: %s %s\n", message->topic, (char *)message->payload);
        printf("output_topic: %s %s\n", output_topic, plainP);
        // 释放内存
        free(base64_plainS);
        free(plainP);
        free(base64_plainP);
    } else {
        printf("%s (null)\n", message->topic);
    }
}

typedef struct demo_alg_kv_s {
    isec_cipher_alg_e sym;
    const char *key;
    const char *iv;
} demo_alg_kv_t;

/* 设置算法列表 */
const demo_alg_kv_t  algs[] = {
    { isec_cipher_alg_sm4,    "U0OSZpZ28sV60VyodNkcdg==" ,"Sdfv8xjtEQcTvwDXevDLtA=="},
    { isec_cipher_alg_des,    "pD4LiUyGpF0=" ,"RgK70OrYvQE="},
    { isec_cipher_alg_3des_112,   "6fF/5VsZJc5owkrlW3z9MQ==" ,"RgK70OrYvQE="},
    { isec_cipher_alg_3des_168, "bo9rYmtKrUpkx/hhdtoNrU8QZ8d/kQ6z" ,"RgK70OrYvQE="},
    { isec_cipher_alg_aes_128, "4zqMaAY/J+Haa+h8WA7x1w==" ,"Sdfv8xjtEQcTvwDXevDLtA=="},
    { isec_cipher_alg_aes_192, "NztNehnKUuKu0TiTKUi5BQ71l/oIpUe3" ,"Sdfv8xjtEQcTvwDXevDLtA=="},
    { isec_cipher_alg_aes_256, "1E2oIWh4gL8WnFx4ZGU4JgTaUb5gim4azfNzz78Jbbg=" ,"Sdfv8xjtEQcTvwDXevDLtA=="}
};

int main(int argc, char *argv[])
{
    /*
        // 打印程序的名称
        printf("Program name: %s\n", argv[0]);

        // 打印命令行参数的数量
        printf("Number of arguments: %d\n", argc - 1);

        // 打印所有命令行参数的值
        printf("Arguments:\n");
        int i;
        for (i = 1; i < argc; i++) {
            printf("%d: %s\n", i, argv[i]);
        }
    */

    isec_global_init();
    // isec_global_set_logger("stderr");
    // rv = demo_main(argc, argv);
    struct mosquitto *mosq = NULL;
    // const char *host = "119.96.80.54";
    // int port = 51883;
    // const char *topic = "test";

    const char *host = NULL;
    int port = 0;
    const char *input_topic = NULL;

    // 遍历命令行参数
    int i;
    for (i = 1; i < argc; i++) {
         // 查找 -ip=
        if (strncmp(argv[i], "-host=", strlen("-host=")) == 0) {
            // 提取值
            host = argv[i] + strlen("-host=");
            printf("Host: %s\n", host);
        }
         // 查找 -port=
        if (strncmp(argv[i], "-port=", strlen("-port=")) == 0) {
            // 提取值
             const char *port_value = argv[i] + strlen("-port=");

            printf("Port: %s\n", port_value);
            port = atoi(port_value);
        }
        // 查找 -input-topic=
        if (strncmp(argv[i], "-input-topic=", strlen("-input-topic=")) == 0) {
            // 提取值
            input_topic = argv[i] + strlen("-input-topic=");
            printf("Input Topic: %s\n", input_topic);
        }

        // 查找 -output-topic=
        if (strncmp(argv[i], "-output-topic=", strlen("-output-topic=")) == 0) {
            // 提取值
            output_topic = argv[i] + strlen("-output-topic=");
            printf("Output Topic: %s\n", output_topic);
        }
    }
    if (host == NULL)
    {
        printf("No Host\n");
        return 0;
    }

    if (port == 0)
    {
        printf("No Port\n");
        return 0;
    }

    if (input_topic == NULL)
    {
        printf("No Input Topic\n");
        return 0;
    }

    if (output_topic == NULL)
    {
        printf("No Output Topic\n");
        return 0;
    }

    int keepalive = 60;

    mosquitto_lib_init();
    mosq = mosquitto_new(NULL, true, NULL);
    
    if(mosq){
        mosquitto_message_callback_set(mosq, on_message);

        if(mosquitto_connect(mosq, host, port, keepalive)){
            fprintf(stderr, "Unable to connect.\n");
            return 1;
        }

        mosquitto_subscribe(mosq, NULL, input_topic, 0);

        while(1){
            if(mosquitto_loop(mosq, -1, 1) != MOSQ_ERR_SUCCESS){
                fprintf(stderr, "Connection error.\n");
                return 1;
            }
        }
    
        mosquitto_destroy(mosq);
    }
    
    mosquitto_lib_cleanup();

    return 0;
}

static int
demo_main(char *base64_dataT, char *base64_plainP, char * plainP)
{
    int                    rv                   = isec_err_fail;
    /* 对称算法配置 */
    isec_cipher_config_t   cipher_config;
    /* 算法类型 */
    isec_cipher_alg_e      alg_mode             = isec_cipher_alg_none;
    /* 算法模式，ecb ，cbc，cfb, ofb, ctr, gcm  */
    isec_cipher_mode_e     mode                 = isec_cipher_mode_none;
    /* 填充模式：0：不填充，    1：P7/P5填充 */
    isec_cipher_padding_e  padding              = isec_cipher_padding_pkcs7;
    /* 初始化明文空间 */
    unsigned char          *data                = NULL;
    int                    data_len             =  0;
    /* 对称密钥 */
    unsigned char          *key                 = NULL;
    int                    key_len              =  0;
    /* 初始化向量 */
    unsigned char          *iv                  = NULL;
    int                    iv_len               =  0;
    /* 初始化aad */
    unsigned char          *aad                 = NULL;
    int                    aad_len              =  0;
    /* 初始化tag */
    unsigned char          *tag                 = NULL;
    int                    tag_len              =  0;
    /* 初始化明文空间，明文缓冲区大小=明文长度+ISEC_MAX_KEY_LEN */
    unsigned char          *plain               = NULL;
    int                     plain_len           =  0;
    /* 初始化算法和算法模式 */
    char                   *alg                 = NULL;
    char                   *mode_str            = NULL;
    /* 临时变量 */
    int                     padding_option      =  1;
    char                   *base64_plain        = NULL;
    int                     baes64_plain_len    =  0;
    char                   *base64_tag          = NULL;
    int                     base64_tag_len      = 0;
    char                   *base64_data         = NULL;
    int                     base64_data_len     = 0;
    char                   *base64_key          = NULL;
    int                     base64_key_len      = 0;
    char                   *base64_iv           = NULL;
    int                     base64_iv_len       = 0;
    char                   *base64_aad          = NULL;
    int                     base64_aad_len      = 0;

    // (void)argc, (void)argv;

    /* 1.接收用户输入算法并解析算法 */
    {
        /* 1.1 接收用户算法 */
        // printf("Input Algorithm:\n");
        // printf("  Algorithm     : (Example: 'sm4','des','3des_112','3des_168','aes_128','aes_192','aes_256')\n");
        if (NULL == ( alg = "sm4")) {
            demo_log_error("demo scanf error");
            rv = isec_err_fail; goto end;
        }

        // alg = "sm4";
        /* 1.2 解析用户算法 */
        if (0 == strcmp((const char*)alg, (const char*)"sm4")) {
            alg_mode = isec_cipher_alg_sm4;
        } else if(0 == strcmp((const char*) alg, (const char*) "des")) {
            alg_mode = isec_cipher_alg_des;
        } else if (0 == strcmp( (const char*) alg, (const char*) "3des_112")) {
            alg_mode = isec_cipher_alg_3des_112;
        } else if (0 == strcmp( (const char*) alg, (const char*) "3des_168")) {
            alg_mode = isec_cipher_alg_3des_168;
        } else if (0 == strcmp( (const char*) alg, (const char*) "aes_128")) {
            alg_mode = isec_cipher_alg_aes_128;
        } else if (0 == strcmp( (const char*) alg, (const char*) "aes_192")) {
            alg_mode = isec_cipher_alg_aes_192;
        } else if (0 == strcmp( (const char*) alg, (const char*) "aes_256")) {
            alg_mode = isec_cipher_alg_aes_256;
        } else {
            demo_log_error("Algorithm does not exist!!\n");
            rv = isec_err_param_invalid; goto end;
        }
    }

    /* 2.输入算法模式并解析模式 */
    {
        /* 2.1 输入算法模式 */
        // printf("%s-Decrypt:\n", alg);
        // printf("  Mode          : (Example: 'ecb', 'cbc', 'cfb', 'ofb', 'ctr', 'gcm')\n");
        if (NULL == (mode_str = "cbc")) {
            demo_log_error("demo scanf error");
            rv = isec_err_fail; goto end;
        }
        // mode_str = "cbc";
        /* 2.2 解析算法模式 */
        if (0 == strcmp((const char*) mode_str, (const char*)"ecb")) {
            mode = isec_cipher_mode_ecb;
        } else if (0 == strcmp((const char*) mode_str, (const char*) "cbc")) {
            mode = isec_cipher_mode_cbc;
        } else if (0 == strcmp((const char*) mode_str, (const char*) "cfb")) {
            mode = isec_cipher_mode_cfb;
        } else if (0 == strcmp((const char*) mode_str, (const char*) "ofb")) {
            mode = isec_cipher_mode_ofb;
        } else if (0 == strcmp((const char*) mode_str, (const char*) "ctr")) {
            mode = isec_cipher_mode_ctr;
        } else if (0 == strcmp((const char*) mode_str, (const char*) "gcm")) {
            mode = isec_cipher_mode_gcm;
        } else {
            demo_log_error("Not Such Mode!!!\n");
            rv = isec_err_param_invalid; goto end;
        }
    }

    /* 3.输入密钥并解析密钥 */
    {
        /* 3.1 输入密钥 */
        // printf("  Key(Base64)   : (Example: '%s')\n",algs[alg_mode - 1].key);
        if (NULL == (base64_key = "U0OSZpZ28sV60VyodNkcdg==")) {
            demo_log_error("demo scanf error");
            rv = isec_err_fail; goto end;
        }
        // base64_key = "U0OSZpZ28sV60VyodNkcdg==";
        base64_key_len = (int)strlen(base64_key);

        /* 3.2 获取解析后密钥数据长度 */
        if (isec_err_none != (rv = isec_crypto_base64_decode(base64_key, base64_key_len, NULL, &key_len))) {
            demo_log_error("base64(%p: %d) decode data(%d) fail", base64_key, base64_key_len, key_len);
            goto end;
        }

        /* 3.3 分配密钥缓冲区大小 */
        key_len += 1;
        if (NULL == (key = (unsigned char *)malloc((size_t)key_len))) {
            demo_log_error("malloc fail");
            rv = isec_err_memory; goto end;
        }

        /* 3.4 获取解析后密钥数据 */
        if (isec_err_none != (rv = isec_crypto_base64_decode(base64_key, base64_key_len, key, &key_len))) {
            demo_log_error("base64(%p: %d) decode data(%p: %d) fail", base64_key, base64_key_len, key, key_len);
            goto end;
        }
    }

    /* 4.输入向量并解析向量 */
    {
        if (mode != isec_cipher_mode_ecb) {
            /* 4.1 输入向量 */
            // printf("  IV(Base64)    : (Example: '%s')\n",algs[alg_mode - 1].iv);
            if (NULL == (base64_iv = "Sdfv8xjtEQcTvwDXevDLtA==")) {
                demo_log_error("demo scanf error");
                rv = isec_err_fail; goto end;
            }
            // base64_iv = "Sdfv8xjtEQcTvwDXevDLtA==";
            base64_iv_len = (int)strlen(base64_iv);

            /* 4.2 获取base64解码后向量长度 */
            if (isec_err_none != (rv = isec_crypto_base64_decode(base64_iv, base64_iv_len, NULL, &iv_len))) {
                demo_log_error("base64(%p: %d) decode data(%d) error", base64_iv, base64_iv_len, iv_len);
                goto end;
            }

            /* 4.3 分配解码后向量长度 */
            iv_len += 1;
            if (NULL == (iv = malloc((size_t)iv_len))) {
                demo_log_error("malloc fail");
                rv = isec_err_memory; goto end;
            }

            /* 4.4 获取base64解码后的向量数据 */
            if (isec_err_none != (rv = isec_crypto_base64_decode(base64_iv, base64_iv_len, iv, &iv_len))) {
                demo_log_error("base64(%p: %d) decode data(%p: %d) error", base64_iv, base64_iv_len, iv, iv_len);
                goto end;
            }

        }

    }

    /* 5.输入解密数据并解析解密数据 */
    {
        /* 5.1 输入密文数据 */
        if (NULL == (base64_data = base64_dataT)) {
            demo_log_error("demo scanf error");
            rv = isec_err_fail; goto end;
        }
        // base64_data = base64_dataT;
        base64_data_len = (int)strlen(base64_data);

        /* 5.2 获取解码后密文数据长度 */
        if (isec_err_none != (rv = isec_crypto_base64_decode(base64_data, base64_data_len, NULL, &data_len))) {
            demo_log_error("base64(%p: %d) decode data(%d) fail", base64_data, base64_data_len, data_len);
            goto end;
        }

        /* 5.3 分配解码后密文缓冲区大小 */
        data_len += 1;
        if (NULL == (data = malloc((size_t)data_len))) {
            demo_log_error("malloc fail");
            rv = isec_err_memory; goto end;
        }

        /* 5.4 获取解码后密文数据 */
        if (isec_err_none != (rv = isec_crypto_base64_decode(base64_data, base64_data_len, data, &data_len))) {
            demo_log_error("base64(%p: %d) decode data(%p: %d) fail", base64_data, base64_data_len, data, data_len);
            goto end;
        }
    }

    /* 6.设置填充模式 */
    // printf("  Padding       : (Exapmpe: '0' - none padding, '1' - pkcs5/pkcs7 padding)\n");
    // printf("  Padding       : "); scanf("%d", &padding_option);
    padding = padding_option ? isec_cipher_padding_pkcs7 : isec_cipher_padding_none;

    // /* 7.补充特殊算法所需参数 */
    // {
    //     /* 7.1 如果是GCM算法，需要额外设置AAD附加数据和TAG认证数据 */
    //     if (isec_cipher_mode_gcm == mode) {
    //         /* 7.1.1 输入的附加数据 */
    //         getchar();
    //         if (NULL == (base64_aad = demo_scanf("  AAD(Base64)   : "))) {
    //             demo_log_error("demo scanf error");
    //             rv = isec_err_fail; goto end;
    //         }
    //         base64_aad_len = (int)strlen(base64_aad);

    //         /* 7.1.2 获取解码后附加数据长度 */
    //         if (isec_err_none != (rv = isec_crypto_base64_decode(base64_aad, base64_aad_len, NULL, &aad_len))) {
    //             demo_log_error("base64(%p: %d) decode data(%d)", base64_aad, base64_aad_len, aad_len);
    //             goto end;
    //         }

    //         /* 7.1.3 分配解码后附加数据缓冲区大小 */
    //         aad_len += 1;
    //         if (NULL == (aad = malloc((size_t)aad_len))) {
    //             demo_log_error("malloc fail");
    //             rv = isec_err_memory; goto end;
    //         }

    //         /* 7.1.4 获取解码后附加数据 */
    //         if (isec_err_none != (rv = isec_crypto_base64_decode(base64_aad, base64_aad_len, aad, &aad_len))) {
    //             demo_log_error("base64(%p: %d) decode data(%p: %d)", base64_aad, base64_aad_len, aad, aad_len);
    //             goto end;
    //         }

    //         /* 7.1.5 输入认证数据 */
    //         if (NULL == (base64_tag = demo_scanf("  TAG(Base64)   : "))) {
    //             demo_log_error("demo scanf fail");
    //             rv = isec_err_fail; goto end;
    //         }
    //         base64_tag_len = (int)strlen(base64_tag);

    //         /* 7.1.6 获取解码后认证数据长度 */
    //         if (isec_err_none != (rv = isec_crypto_base64_decode(base64_tag, base64_tag_len, NULL, &tag_len))) {
    //             demo_log_error("base64(%p: %d) decoed data(%d)", base64_tag, base64_tag_len, tag_len);
    //             goto end;
    //         }

    //         /* 7.1.7 分配认证数据缓冲区 */
    //         tag_len += 1;
    //         if (NULL == (tag = malloc((size_t)tag_len))) {
    //             demo_log_error("malloc fail");
    //             rv = isec_err_memory; goto end;
    //         }

    //         /* 7.1.8 获取解码后认证数据 */
    //         if (isec_err_none != (rv = isec_crypto_base64_decode(base64_tag, base64_tag_len, tag, &tag_len))) {
    //             demo_log_error("base64(%p: %d) decoed data(%p: %d)", base64_tag, base64_tag_len, tag, tag_len);
    //             goto end;
    //         }
    //     }
    // }

    /* 8.配置对称算法 */
    memset(&cipher_config, 0, sizeof(cipher_config));
    cipher_config.alg      = alg_mode; /* 必选，设置算法类别 */
    cipher_config.mode     = mode;     /* 必选，对称算法模式 */
    cipher_config.padding  = padding;  /* 对称算法填充 */
    cipher_config.key      = key;      /* 对称密钥 */
    cipher_config.key_len  = key_len;  /* 对称密钥长度 */
    cipher_config.iv       = iv;       /* 可选，初始化向量。若对称算法模式非ECB，则必须传入该参数 */
    cipher_config.iv_len   = iv_len;   /* 初始化向量长度 */

    cipher_config.aad      = aad;      /* 可选，认证数据，仅GCM算法需要配置 */
    cipher_config.aad_len  = aad_len;  /* 可选，认证数据，仅GCM算法需要配置 */

    /* 9.对用户输入数据进行解密 */
    {
        /* 9.1 获取解密后明文数据长度 */
        if (isec_err_none != (rv = isec_crypto_decrypt(&cipher_config, data, data_len, tag, tag_len , NULL, &plain_len))) {
            demo_log_error("decrypt fail, alg(%s), alg_mode(%s), padding(%d), key(%p), iv(%p), data(%p: %d), plain(%d)",
                alg, mode_str, padding, key, iv, data, data_len, plain_len);
            goto end;
        }

        /* 9.2 分配解密后明文缓冲区大小 */
        plain_len += 1;
        if (NULL == (plain = (unsigned char *)malloc((size_t)plain_len))) {
            demo_log_error("malloc fail");
            rv = isec_err_memory; goto end;
        }

        /* 9.3 获取解密后明文数据 */
        if (isec_err_none != (rv = isec_crypto_decrypt(&cipher_config, data, data_len, tag, tag_len , plain, &plain_len))) {
            demo_log_error("decrypt fail, alg(%s), alg_mode(%s), padding(%d), key(%p), iv(%p), data(%p: %d), plain(%p: %d)",
                alg, mode_str, padding, key, iv, data, data_len, plain, plain_len);
            goto end;
        }
        // 
        memcpy(plainP,plain,plain_len);
    }

    /* 10.对解密后的明文进行base64编码后输出 */
    {
        /* 10.1 获取base64编码后的明文长度 */
        if (isec_err_none != (rv = isec_crypto_base64_encode(plain, plain_len, NULL, &baes64_plain_len))) {
            demo_log_error("data(%p: %d) encode base64(%d) fail", plain, plain_len, baes64_plain_len);
                goto end;
        }

        /* 10.2 分配编码后的明文缓冲区大小 */
        baes64_plain_len += 1;
        if (NULL == (base64_plain = malloc((size_t)baes64_plain_len))) {
            demo_log_error("malloc fail");
            rv = isec_err_memory; goto end;
        }

        /* 10.3 获取base64编码后的明文数据 */
        if (isec_err_none != (rv = isec_crypto_base64_encode(plain, plain_len, base64_plain, &baes64_plain_len))) {
            demo_log_error("data(%p: %d) encode base64(%p: %d) fail", plain, plain_len, base64_plain, baes64_plain_len);
                goto end;
        }
        memcpy(base64_plainP,base64_plain,baes64_plain_len);
    }
    /* 成功 */
    rv = isec_err_none;
    printf("  Plain         : %s \n", plainP);
    printf("  base64_Plain         : %s \n", base64_plainP);
    end:
        // if (mode_str)       free(mode_str);
        // if (alg)            free(alg);
        if (base64_plain)   free(base64_plain);
        if (plain)          free(plain);
        // if (base64_aad)     free(base64_aad);
        // if (aad)            free(aad);
        // if (base64_tag)     free(base64_tag);
        // if (tag)            free(tag);
        if (key)            free(key);
        // if (base64_key)     free(base64_key);
        if (iv)             free(iv);
        // if (base64_iv)      free(base64_iv);

        if (data)           free(data);
        // if (base64_data)    free(base64_data);

    return rv;
}

static char *
demo_scanf(const char *fmt, ...)
{
    va_list  ap;
    char    *s = NULL;
    size_t   slen = 0;
    int      ch = 0;

    /* print */
    va_start(ap, fmt);
    vprintf(fmt, ap);
    va_end(ap);

    /* input */
    while ('\n' != (ch = getchar())) {
        /* fix: here not check realloc return value */
        s = (NULL == s) ? malloc((size_t)2) : realloc(s, slen + 2);
        s[slen++] = (char)ch;
        s[slen] = '\0';
    }
    return s;
}

